
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _RECTANGLE_HH_
#define _RECTANGLE_HH_

#include "math.hh"
#include "vector.hh"


namespace dc {

//
// A rectangle is a 2d shape defined by point (origin) and size (width & height)
// vectors.  The usual rectangle maths are available.  Valid rectangles have
// positive width and height.  Invalid rectangles will be returned for invalid
// operations such as trying to get the intersection of two rectangles which
// don't intersect.  Using an invalid rectangle is undefined, but it may be used
// after "fixing" it using the fix method which will cause the size to be
// positive and adjust the origin to be the bottom-left point.  Check the
// validity of a rectangle using the isValid method especially if the rectangle
// is a result of an operation which could fail.
//

class rectangle
{
public:
	// Constructors.
	rectangle() {}
	rectangle(point o, point s) {
		origin = o;
		size = s;
	}

	// Rectangle validity.
	bool isValid() const { return size.x >= 0.0 && size.y >= 0.0; }
	void fix()
	{
		if (size.x <= 0.0)
		{
			size.x = 0.0 - size.x;
			origin.x -= size.x;
		}
		if (size.y <= 0.0)
		{
			size.y = 0.0 - size.y;
			origin.y -= size.y;
		}
	}

	// Rectangle operations.
	scalar area() const { return size.x * size.y; }
	scalar perimeter() const {
		return  size.x + size.x + size.y + size.y;
	}

	bool containsPoint(point pt) const
	{
		return origin.x <= pt.x && origin.y <= pt.y &&
			   pt.x <= origin.x + size.x && pt.y <= origin.y + size.y;
	}

	void inset(scalar h, scalar v)
	{
		origin.x	+= h;
		origin.y	+= v;
		size.x		-= h + h;
		size.y		-= v + v;
	}
	void offset(scalar h, scalar v)
	{
		origin.x	-= h;
		origin.y	-= v;
		size.x		+= h + h;
		size.y		+= v + v;
	}

	bool containsRect(const rectangle& r) const;
	bool intersectsRect(const rectangle& r) const;
	rectangle unionWith(const rectangle& r) const;
	rectangle intersectionWith(const rectangle& r) const;

	// Checking equality.
	bool operator == (const rectangle& r) const {
		return origin == r.origin && size == r.size;
	}
	bool operator != (const rectangle& r) const {
		return !(*this == r);
	}

	// Data.
	point origin, size;
};

} // namespace dc


#endif // _RECTANGLE_HH_

