
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <cstdlib>

#include <boost/bind.hpp>

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>

#include "dispatcher.hh"
#include "opengl.hh"

#include "texture.hh"


namespace dc {


class texture_impl
{
	void unloadFromGL()
	{
		if (object)
		{
			glDeleteTextures(1, &object);
			object = 0;
		}
	}

	void contextRecreated(const notification& note)
	{
		unloadFromGL();
		uploadToGL();
	}

public:
	texture_impl(texture* outside, bool keepInMemory)
		: interface(outside), keepData(keepInMemory), object(0), imageData(0)
	{
		dispatcher::instance().addHandler("video.context_recreated",
				boost::bind(&texture_impl::contextRecreated, this, _1),
				this);
	}

	~texture_impl()
	{
		dispatcher::instance().removeHandler(this);

		if (imageData)
		{
			SDL_FreeSurface(imageData);
		}
		unloadFromGL();
	}


	static int powerOfTwo(int input)
	{
		int value = 1;

		while (value < input)
		{
			value <<= 1;
		}
		return value;
	}

	static SDL_Surface* prepareImageForGL(SDL_Surface* surface)
	{
		// Adapted from some public domain code.  This stuff is common enough
		// that it really should be included in SDL_image...  We need this
		// because images loaded with SDL_image aren't exactly OpenGL-ready
		// right out of the box.
		
		int w = powerOfTwo(surface->w);
		int h = powerOfTwo(surface->h);

		// 1. OpenGL images must (generally) have dimensions of a power-of-two.

		SDL_Surface* image = SDL_CreateRGBSurface
		(
			SDL_SWSURFACE,
			w, h,
			32,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
			0x000000FF, 
			0x0000FF00, 
			0x00FF0000, 
			0xFF000000
#else
			0xFF000000,
			0x00FF0000, 
			0x0000FF00, 
			0x000000FF
#endif
		);

		if (!image)
		{
			return 0;
		}

		Uint32 savedFlags = surface->flags&(SDL_SRCALPHA|SDL_RLEACCELOK);
		Uint8  savedAlpha = surface->format->alpha;
		if (savedFlags & SDL_SRCALPHA)
		{
			SDL_SetAlpha(surface, 0, 0);
		}

		SDL_Rect srcArea, destArea;
		srcArea.x = 0; destArea.x = 0;
		srcArea.y = 0; destArea.y = h - surface->h;
		srcArea.w = surface->w;
		srcArea.h = surface->h;
		SDL_BlitSurface(surface, &srcArea, image, &destArea);

		if (savedFlags & SDL_SRCALPHA)
		{
			SDL_SetAlpha(surface, savedFlags, savedAlpha);
		}

		// 2. OpenGL textures make more sense when they are "upside down."

		Uint8 line[image->pitch];

		Uint8 *pixels = static_cast<Uint8*>(image->pixels);
		Uint16 pitch = image->pitch;
		int ybegin = 0;
		int yend = image->h - 1;

		if (SDL_MUSTLOCK(image)) SDL_LockSurface(image);
		while (ybegin < yend)
		{
			memcpy(line, pixels + pitch * ybegin, pitch);
			memcpy(pixels + pitch * ybegin, pixels + pitch * yend, pitch);
			memcpy(pixels + pitch * yend, line, pitch);
			ybegin++;
			yend--;
		}
		if (SDL_MUSTLOCK(image)) SDL_UnlockSurface(image);

		return image;
	}


	void loadImageData()
	{
		SDL_Surface* surface;

		surface = IMG_Load(interface->getPathToFile().c_str());

		if (!surface)
		{
			throw texture::exception("loading failed");
		}

		imageData = prepareImageForGL(surface);
		SDL_FreeSurface(surface);

		if (!imageData)
		{
			throw texture::exception("");
		}

		if (imageData->format->BytesPerPixel == 3)
		{
			mode = GL_RGB;
		}
		else if (imageData->format->BytesPerPixel == 4)
		{
			mode = GL_RGBA;
		}
		else
		{
			SDL_FreeSurface(imageData);
			throw texture::exception("image is not the required 24 or 32 bpp");
		}

		width = imageData->w;
		height = imageData->h;
	}


	void uploadToGL()
	{
		if (object)
		{
			// Already loaded.
			return;
		}

		if (!imageData)
		{
			loadImageData();
		}

		glGenTextures(1, &object);

		glBindTexture(GL_TEXTURE_2D, object);

		glTexImage2D
		(
			GL_TEXTURE_2D,
			0,
			mode,
			imageData->w,
			imageData->h,
			0,
			mode,
			GL_UNSIGNED_BYTE,
			imageData->pixels
		);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

		if (!keepData)
		{
			SDL_FreeSurface(imageData);
			imageData = 0;
		}
	}


	unsigned width;
	unsigned height;
	int mode;
	GLuint object;

	bool keepData;
	SDL_Surface* imageData;

	texture* interface;
};


texture::texture(const std::string& name, bool keepInMemory)
	: resource(name), impl(new texture_impl(this, keepInMemory)) {}


void texture::bind()
{
	glBindTexture(GL_TEXTURE_2D, getObject());
}

GLuint texture::getObject()
{
	if (!impl->object)
	{
		impl->uploadToGL();
	}

	return impl->object;
}


unsigned texture::getWidth()
{
	if (!impl->object)
	{
		impl->uploadToGL();
	}

	return impl->width;
}

unsigned texture::getHeight()
{
	if (!impl->object)
	{
		impl->uploadToGL();
	}

	return impl->height;
}


} // namespace dc

